"""The core abstraction for hardware in PyRTL.

Included in this file you will find:

* `LogicNet` -- the core class holding a "net" in the netlist
* `Block` -- a collection of nets with associated access and error checking
* `working_block` -- the "current" Block to which, by default, all created nets are
                     added
* `modes` -- access methods for "modes" such as debug

"""

from __future__ import annotations

import collections
import inspect
import keyword
import re
from typing import TYPE_CHECKING, NamedTuple

from pyrtl.pyrtlexceptions import PyrtlError, PyrtlInternalError

if TYPE_CHECKING:
    from pyrtl.memory import MemBlock
    from pyrtl.wire import Register, WireVector


# -----------------------------------------------------------------
#    __        __   __
#   |__) |    /  \ /  ` |__/
#   |__) |___ \__/ \__, |  \
#
class LogicNet(NamedTuple):
    """The basic immutable datatype for storing a "net" in a netlist.

    ``LogicNet`` is part of PyRTL's internal representation. What it is, and how it is
    used are only required for advanced PyRTL users.

    A "net" is a Python representation of a hardware logic operation. These operations
    include binary operations like `and` `or` and `not`, arithmetic operations like `+`
    and `-`, and other operations like concatenating wires, splitting wires,
    reading/writing memory, and register logic:

    ===== ========== ========== ======== ====
    op    op_param   args       dests
    ===== ========== ========== ======== ====
    ``&`` ``None``   ``a1, a2`` ``out``  AND two wires together, put result
                                         into ``out``
    ``|`` ``None``   ``a1, a2`` ``out``  OR two wires together, put result into
                                         ``out``
    ``^`` ``None``   ``a1, a2`` ``out``  XOR two wires together, put result
                                         into ``out``
    ``n`` ``None``   ``a1, a2`` ``out``  NAND two wires together, put result
                                         into ``out``
    ``~`` ``None``   ``a1``     ``out``  invert one wire, put result into
                                         ``out``
    ``+`` ``None``   ``a1, a2`` ``out``  add ``a1`` and ``a2``, put result into
                                         ``out``

                                         ``len(out) == max(len(a1), len(a2)) + 1``

                                         Performs *unsigned* addition. Use
                                         :func:`signed_add` for signed
                                         addition.
    ``-`` ``None``   ``a1, a2`` ``out``  subtract ``a2`` from ``a1``, put
                                         result into ``out``

                                         ``len(out) == max(len(a1), len(a2)) + 1``

                                         Performs *unsigned* subtraction. Use
                                         :func:`signed_sub` for signed
                                         subtraction.
    ``*`` ``None``   ``a1, a2`` ``out``  multiply ``a1`` & ``a2``, put result
                                         into ``out``

                                         ``len(out) == len(a1) + len(a2)``

                                         Performs *unsigned* multiplication.
                                         Use :func:`signed_mult` for signed
                                         multiplication.
    ``=`` ``None``   ``a1, a2`` ``out``  check ``a1`` & ``a2`` equal, put
                                         result into ``out`` (0 | 1)
    ``<`` ``None``   ``a1, a2`` ``out``  check ``a1`` less than ``a2``, put
                                         result into ``out``. ``out`` has
                                         bitwidth 1.

                                         Performs *unsigned* comparison. Use
                                         :func:`signed_lt` for signed less
                                         than.
    ``>`` ``None``   ``a1, a2`` ``out``  check ``a1`` greater than ``a2``, put
                                         result into ``out`` ``out`` has
                                         bitwidth 1.

                                         Performs *unsigned* comparison. Use
                                         :func:`signed_gt` for signed greater
                                         than.
    ``w`` ``None``   ``w1``     ``w2``   connects ``w1`` to ``w2``

                                         directional wire with no logical
                                         operation
    ``x`` ``None``   ``x``,     ``out``  multiplexer:

                                         when ``x`` == 0 connect ``a1`` to
                                         ``out``

                                         when ``x`` == 1 connect ``a2`` to
                                         ``out``

                                         ``x`` must be 1-bit and ``len(a1) == len(a2)``
                     ``a1, a2``
    ``c`` ``None``   ``*args``  ``out``  concatenates ``*args`` (wires) into
                                         single WireVector

                                         puts first arg at MSB, last arg at LSB
    ``s`` ``sel``    ``wire``   ``out``  selects bits from wire based on
                                         ``sel`` (slicing syntax)

                                         puts selected bits into ``out``
    ``r`` ``None``   ``next``   ``r1``   on positive clock edge: copies
                                         ``next`` to ``r1``
    ``m`` ``memid``, ``addr``   ``data`` read address addr of mem (with id
                                         ``memid``), put it into ``data``

          ``mem``
    ``@`` ``memid``, ``addr``            write data to mem (with id ``memid``)
                                         at address ``addr``

                                         request write enable (``wr_en``)

          ``mem``    ``data``,

                     ``wr_en``
    ===== ========== ========== ======== ====

    """

    op: str
    """Operation performed by the ``LogicNet``."""
    op_param: tuple
    """Static parameters for the operation."""
    args: tuple[WireVector]
    """Input arguments to the operation."""
    dests: tuple[WireVector]
    """Output of the operation."""

    def __str__(self):
        rhs = ", ".join(str(x) for x in self.args)
        lhs = ", ".join(str(x) for x in self.dests)
        options = "" if self.op_param is None else "(" + str(self.op_param) + ")"

        from pyrtl.helperfuncs import _currently_in_jupyter_notebook

        if _currently_in_jupyter_notebook():
            # Output the working block as a Latex table
            # Escape all Underscores
            rhs = rhs.replace("_", "\\_")
            lhs = lhs.replace("_", "\\_")
            options = options.replace("_", "\\_")
            if self.op in "&|":
                return f"{lhs} & \\leftarrow \\{self.op} \\, - & {rhs} {options} \\\\"
            if self.op in "wn+-*<>xcsr":
                return f"{lhs} & \\leftarrow {self.op} \\, - & {rhs} {options} \\\\"
            if self.op in "=":
                return f"{lhs} & \\leftarrow \\, {self.op} \\, - & {rhs} {options} \\\\"
            if self.op in "^":
                return f"{lhs} & \\leftarrow \\oplus \\, - & {rhs} {options} \\\\"
            if self.op in "~":
                return f"{lhs} & \\leftarrow \\sim \\, - & {rhs} {options} \\\\"

            if self.op in "m@":
                memid, memblock = self.op_param
                extrainfo = "memid=" + str(memid)
                extrainfo = extrainfo.replace("_", "\\_")
                name = memblock.name
                name = name.replace("_", "\\_")
                if self.op == "m":
                    return (
                        f"{lhs} & \\leftarrow m \\, - &  "
                        f"{name}[{rhs}]({extrainfo}) \\\\"
                    )
                addr, data, we = (str(x) for x in self.args)
                addr = addr.replace("_", "\\_")
                data = data.replace("_", "\\_")
                we = we.replace("_", "\\_")
                return (
                    f"{name}[{addr}] & \\leftarrow @ \\, - & "
                    f"{data} we={we} ({extrainfo}) \\\\"
                )
            msg = f'error, unknown op "{self.op}"'
            raise PyrtlInternalError(msg)

        # not in ipython
        if self.op in "w~&|^n+-*<>=xcsr":
            options = " " + options if options else ""
            return f"{lhs} <-- {self.op} -- {rhs}{options}"
        if self.op in "m@":
            memid, memblock = self.op_param
            extrainfo = "memid=" + str(memid)
            if self.op == "m":
                return f"{lhs} <-- m --  {memblock.name}[{rhs}]({extrainfo})"
            addr, data, we = (str(x) for x in self.args)
            return f"{memblock.name}[{addr}] <-- @ -- {data} we={we} ({extrainfo})"
        msg = f'error, unknown op "{self.op}"'
        raise PyrtlInternalError(msg)

    def __hash__(self):
        # it seems that namedtuple is not always hashable
        return hash(tuple(self))

    def __eq__(self, other):
        # We can't be going and calling __eq__ recursively on the logic nets for all of
        # the args and dests because that will actually *create* new logic nets which is
        # very much not what people would expect to happen. Instead we define equality
        # as the immutable fields being equal and the list of args and dests being
        # references to the same objects.
        return (
            self.op == other.op
            and self.op_param == other.op_param
            and len(self.args) == len(other.args)
            and len(self.dests) == len(other.dests)
            and all(self.args[i] is other.args[i] for i in range(len(self.args)))
            and all(self.dests[i] is other.dests[i] for i in range(len(self.dests)))
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    def _compare_error(self, _other):
        """Throw error when LogicNets are compared.

        Comparisons get you in a bad place between while you can compare op and op_param
        safely, the args and dests are references to mutable objects with comparison
        operators overloaded.
        """
        msg = (
            "Greater than and less than comparisons between LogicNets are not supported"
        )
        raise PyrtlError(msg)

    __lt__ = _compare_error
    __gt__ = _compare_error
    __le__ = _compare_error
    __ge__ = _compare_error


class Block:
    """``Block`` encapsulates a `netlist <https://en.wikipedia.org/wiki/Netlist>`_.

    A ``Block`` in PyRTL is the class that stores a netlist and provides basic access
    and error checking members. Each ``Block`` has well defined :class:`Inputs<Input>`
    and :class:`Outputs<Output>`, and contains both the basic logic elements and
    references to the :class:`WireVectors<WireVector>` and :class:`MemBlocks<MemBlock>`
    that connect them together.

    The logic structure is primarily contained in :attr:`~Block.logic` which holds a
    :class:`set` of :class:`LogicNets <LogicNet>`. Each :class:`LogicNet` describes a
    primitive operation (such as an adder or memory), a 4-tuple of:

    1. The :attr:`~LogicNet.op` (a single character describing the operation such as
       ``+`` or ``r``).

    2. A set of static :attr:`~LogicNet.op_param` for the operation, such as the bit
       slices to select for the ``s`` "selection" operation.

    3. A tuple :attr:`~LogicNet.args` containing the :class:`WireVectors<WireVector>`
       connected as inputs to the :class:`LogicNet`.

    4. A tuple :attr:`~LogicNet.dests` containing the :class:`WireVectors<WireVector>`
       connected as outputs from the :class:`LogicNet`.

    Below is a list of the basic operations. These properties (more formally specified)
    should all be checked by :meth:`sanity_check`.

    - Most logical and arithmetic :class:`ops<LogicNet.op>` are pretty self explanatory.
      Each takes exactly two :class:`~LogicNet.args`, and they should perform the
      arithmetic or logical operation specified.

      :class:`ops<LogicNet.op>`: ``&``, ``|``, ``^``, ``n``, ``~``, ``+``, ``-``,
      ``*``.

      All inputs must be the same :attr:`~WireVector.bitwidth`.  Logical operations
      produce as many bits as are in the input, while ``+`` and ``-`` produce ``n + 1``
      bits, and ``*`` produces ``2 * n`` bits.

    - In addition there are some operations for performing comparisons that should
      perform the operation specified. The ``=`` :class:`~LogicNet.op` is checking to
      see if the bits of the :class:`~LogicNet.args` vectors are equal, while ``<`` and
      ``>`` do *unsigned* arithmetic comparison. All comparisons generate a single bit
      :class:`dest<LogicNet.dests>` (``1`` for ``True``, ``0`` for ``False``).

    - The ``w`` :class:`~LogicNet.op` is simply a directional wire that connects
      :class:`~LogicNet.args` to :class:`~LogicNet.dests`. It has no logic function.

    - The ``x`` :class:`~LogicNet.op` is a multiplexer which takes a select bit and two
      signals as :class:`~LogicNet.args`. If the value of the select bit is ``0`` it
      selects the second :class:`arg<LogicNet.args>`; if it is ``1`` it selects the
      third :class:`arg<LogicNet.args>`. Select must be a single bit, while the other
      two :class:`~LogicNet.args` must be the same length.

    - The ``c`` :class:`~LogicNet.op` is the concatenation operator and combines any
      number of :class:`WireVector` :class:`~LogicNet.args` (``a``, ``b``, ..., ``z``)
      into a single new :class:`WireVector` with ``a`` in the MSB and ``z`` (or whatever
      is last) in the LSB position.

    - The ``s`` :class:`~LogicNet.op` is the selection operator and chooses, based on
      the :class:`~LogicNet.op_param` specified, a subset of the logic bits from a
      :class:`WireVector` to select. Repeats are accepted.

    - The ``r`` :class:`~LogicNet.op` is a register and on posedge, simply copies the
      value from :class:`arg<LogicNet.args>` to the register's
      :class:`dest<LogicNet.dests>`.

    - The ``m`` :class:`~LogicNet.op` is a memory block read port, which supports async
      reads (acting like combinational logic). Multiple read (and write) ports are
      possible to the same memory but each ``m`` defines only one of those. The
      :class:`~LogicNet.op_param` is a tuple containing two references: the ``memid``,
      and a reference to the :class:`MemBlock` containing this port. The
      :class:`MemBlock` should only be used for debug and sanity checks. Each read port
      has one ``addr`` (an :class:`arg<LogicNet.args>`) and one ``data`` (a
      :class:`dest<LogicNet.dests>`).

    - The ``@`` (update) :class:`~LogicNet.op` is a memory block write port, which
      supports synchronous writes (writes are "latched" at positive edge). Multiple
      write (and read) ports are possible to the same memory but each ``@`` defines only
      one of those. The :class:`~LogicNet.op_param` is a tuple containing two
      references: the ``memid``, and a reference to the :class:`MemBlock`. Writes have
      three :class:`~LogicNet.args` (``addr``, ``data``, and write enable ``we_en``).
      The :class:`~LogicNet.dests` should be an empty tuple. You will not see a written
      value change until the following cycle. If multiple writes happen to the same
      address in the same cycle the behavior is currently undefined.

    The connecting elements (:class:`~LogicNet.args` and :class:`~LogicNet.dests`)
    should be :class:`WireVectors<WireVector>` or derived from :class:`WireVector`, and
    should be registered with the ``Block`` using :meth:`add_wirevector`.
    :class:`LogicNets<LogicNet>` should be registered using :meth:`add_net`.

    In addition, there is a member :attr:`~Block.legal_ops` which defines the set of
    operations that can be legally added to the ``Block``. By default it is set to all
    of the above defined operations, but it can be useful in certain cases to only allow
    a subset of operations (such as when transforms are being done that are "lowering"
    the ``Block`` to more primitive ops).

    See :class:`LogicNet` for a complete list of defined operations.
    """

    logic: set[LogicNet]
    """Set of :class:`LogicNets<LogicNet>` belonging to this ``Block``."""

    legal_ops: set[str]
    """Set of allowed :attr:`ops<LogicNet.op>` in this block."""

    def __init__(self):
        """Creates an empty hardware block."""
        self.logic = set()  # set of nets, each is a LogicNet named tuple
        self.wirevector_set = set()  # set of all WireVectors
        self.wirevector_by_name = {}  # map from name->WireVector, used for performance
        # pre-synthesis WireVectors to post-synthesis vectors
        self.legal_ops = set("w~&|^n+-*<>=xcsrm@")  # set of legal OPS
        # map from WireVectors -> exceptions, used by rtl_assert
        self.rtl_assert_dict = {}
        # map from name->memblock, for easy access to memblock objs
        self.memblock_by_name = {}

    def __str__(self):
        """String form has one LogicNet per line."""
        from pyrtl.helperfuncs import (
            _currently_in_jupyter_notebook,
            _print_netlist_latex,
        )

        if _currently_in_jupyter_notebook():
            _print_netlist_latex(list(self))
            return " "
        return "\n".join(str(net) for net in self)

    def add_wirevector(self, wirevector: WireVector):
        """
        :param wirevector: :class:`WireVector` to add to ``self``.
        """
        self.sanity_check_wirevector(wirevector)
        self.wirevector_set.add(wirevector)
        self.wirevector_by_name[wirevector.name] = wirevector

    def remove_wirevector(self, wirevector: WireVector):
        """
        :param wirevector: :class:`WireVector` to remove from ``self``.
        """
        self.wirevector_set.remove(wirevector)
        del self.wirevector_by_name[wirevector.name]

    def add_net(self, net: LogicNet):
        """Add logic to the ``Block``.

        The passed :class:`LogicNet` is checked and added to the ``Block``. No
        :class:`WireVectors<WireVector>` are added; they must be added seperately with
        :meth:`add_wirevector`.

        :param net: :class:`LogicNet` to add to ``self``.
        """

        self.sanity_check_net(net)
        self.logic.add(net)

    def _add_memblock(self, mem):
        """Registers a memory to the block.

        Note that this is done automatically when a memory block is created and isn't
        intended for use by PyRTL end users. This is so non-local memories can be
        accessed later on (e.g. for instantiating during a simulation).
        """
        self.sanity_check_memblock(mem)
        self.memblock_by_name[mem.name] = mem

    def get_memblock_by_name(self, name: str, strict: bool = False) -> MemBlock:
        """Get a :class:`MemBlock` from the ``Block``, by name.

        .. doctest only::

            >>> import pyrtl
            >>> pyrtl.reset_working_block()

        Useful for getting access to internal :class:`MemBlocks<MemBlock>` for testing.
        For example, the :class:`Simulation` constructor requires a :class:`MemBlock`
        reference for its ``memory_value_map``. Instead of passing the :class:`MemBlock`
        around in your code, you can get a reference to the :class:`MemBlock` from the
        global :func:`working_block`::

            >>> def read_hidden_mem(read_addr: pyrtl.WireVector) -> pyrtl.WireVector:
            ...     mem = pyrtl.MemBlock(bitwidth=8, addrwidth=4, name='hidden_mem')
            ...     return mem[read_addr]

            >>> read_addr = pyrtl.Input(4, 'read_addr')
            >>> data = pyrtl.Output(8, 'data')
            >>> data <<= read_hidden_mem(read_addr)

            >>> hidden_mem = pyrtl.working_block().get_memblock_by_name('hidden_mem')
            >>> sim = pyrtl.Simulation(memory_value_map={hidden_mem: {3: 7}})
            >>> sim.step(provided_inputs={read_addr.name: 3})
            >>> sim.inspect(data.name)
            7

        Returns ``None`` if no matching :class:`MemBlock` can be found. However, if
        ``strict=True``, then this will instead throw a :class:`PyrtlError` when no
        match is found.

        :param name: Name of the :class:`MemBlock` to retrieve.
        :param strict: When ``True``, raises an exception when no match is found.
            Defaults to ``False``.

        :raise PyrtlError: When ``strict=True`` and no match is found.

        :return: The :class:`MemBlock` object with specified name
        """
        if name in self.memblock_by_name:
            return self.memblock_by_name[name]
        if strict:
            msg = f"error, block does not have a memblock named {name}"
            raise PyrtlError(msg)
        return None

    def wirevector_subset(
        self, cls: tuple[type] | None = None, exclude: tuple[type] = ()
    ) -> set[WireVector]:
        """Return a subset of the ``Block's`` :class:`WireVectors<WireVector>`.

        Filters :class:`WireVectors<WireVector>` by type.

        .. doctest only::

            >>> import pyrtl
            >>> pyrtl.reset_working_block()

        If ``cls`` is ``None``, returns all the :class:`WireVectors<WireVector>`
        associated with the ``Block``. If ``cls`` is a single type, or a tuple of types,
        only :class:`WireVectors<WireVector>` of the matching types will be returned.
        This is helpful for getting all of a ``Block``'s :class:`Inputs<Input>`,
        :class:`Outputs<Output>`, or :class:`Registers<Register>` for example::

            >>> len([pyrtl.Input(bitwidth=4) for _ in range(2)])
            2
            >>> len([pyrtl.Output(bitwidth=4) for _ in range(3)])
            3
            >>> block = pyrtl.working_block()

            >>> inputs = block.wirevector_subset(pyrtl.Input)
            >>> len(inputs)
            2
            >>> all(isinstance(input, pyrtl.Input) for input in inputs)
            True

            >>> non_inputs = block.wirevector_subset(exclude=pyrtl.Input)
            >>> len(non_inputs)
            3
            >>> any(isinstance(non_input, pyrtl.Input) for non_input in non_inputs)
            False

        :param cls: Type of :class:`WireVector` objects to include in the returned
            :class:`set`.
        :param exclude: Type of :class:`WireVector` objects to exclude from the returned
            :class:`set`.

        :return: :class:`WireVectors<WireVector>` in the ``Block`` that are both a
                 ``cls`` type and not an ``excluded`` type.
        """
        if cls is None:
            initial_set = self.wirevector_set
        else:
            initial_set = (x for x in self.wirevector_set if isinstance(x, cls))
        if exclude == ():
            return set(initial_set)
        return {x for x in initial_set if not isinstance(x, exclude)}

    def logic_subset(self, op: tuple[str] | None = None) -> set[LogicNet]:
        """Return a subset of the ``Block's`` :class:`LogicNets<LogicNet>`.

        Filters :class:`LogicNets<LogicNet>` by their :attr:`~LogicNet.op`.

        If :class:`~LogicNet.op` is ``None``, returns all the
        :class:`LogicNets<LogicNet>` associated with the ``Block``. Otherwise, returns a
        set of the :class:`LogicNets<LogicNet>` with one of the specified
        :class:`ops<LogicNet.op>`. This is helpful for getting all memories of a block
        for example.

        :param op: :attr:`LogicNet.op` to filter on. Defaults to ``None``.

        :return: :class:`LogicNets<LogicNet>` in the ``Block`` with the corresponding
                 :class:`~LogicNet.op`.
        """
        if op is None:
            return self.logic
        return {x for x in self.logic if x.op in op}

    def get_wirevector_by_name(self, name: str, strict: bool = False) -> WireVector:
        """Return the :class:`WireVector` with matching ``name``.

        :param name: Name of :class:`WireVector`.
        :param strict: If ``True``, raise an exception when no matching
            :class:`WireVector` is found. Defaults to False.

        :raise PyrtlError: if ``strict=True`` and no matching :class:`WireVector` is
            found.

        :return: The :class:`WireVector` with the specified ``name``.
        """
        if name in self.wirevector_by_name:
            return self.wirevector_by_name[name]
        if strict:
            msg = f"error, block does not have a WireVector named {name}"
            raise PyrtlError(msg)
        return None

    class _NetConnectionsDict(dict):
        """Dictionary wrapper for returning the results of the enclosing function.

        User doesn't need to know about this class; it's only for delivering a nice
        error message when _MemIndexed is used as a lookup key.
        """

        def __missing__(self, key):
            from pyrtl.memory import _MemIndexed

            if isinstance(key, _MemIndexed):
                msg = (
                    "Cannot look up a _MemIndexed object's source or destination net. "
                    "Try using its '.wire' attribute as the lookup key instead."
                )
                raise PyrtlError(msg)
            raise KeyError(key)

    def net_connections(
        self, include_virtual_nodes: bool = False
    ) -> tuple[dict[WireVector, LogicNet], dict[WireVector, list[LogicNet]]]:
        """Returns `sources` and `sinks` for each :class:`WireVector` in the ``Block``.

        A :class:`WireVector`'s `source` is the :class:`LogicNet` that sets the
        :class:`WireVector`'s value.

        A :class:`WireVector`'s `sinks` are the :class:`LogicNets<LogicNet>` that use
        the :class:`WireVector`'s value.

        This information helps when building a graph representation for the ``Block``.
        See :func:`net_graph` for an example.

        .. note::

            Consider using :ref:`gate_graphs` instead.

        :param include_virtual_nodes: If ``True``, external `sources` (such as an
            :class:`Inputs<Input>` and :class:`Consts<Const>`) will be represented as
            wires that set themselves, and external `sinks` (such as
            :class:`Outputs<Output>`) will be represented as wires that use themselves.
            If ``False``, these nodes will be excluded from the results.

        :return: Two dictionaries. The first maps :class:`WireVectors<WireVector>` to
                 the :class:`LogicNet` that creates their signal (``wire_src_dict``).
                 The second maps :class:`WireVectors<WireVector>` to a list of
                 :class:`LogicNets<LogicNet>` that use their signal
                 (``wire_sink_dict``).
        """
        src_list = {}
        dst_list = {}

        def add_wire_src(edge, node):
            if edge in src_list:
                msg = (
                    f'Wire "{edge}" has multiple drivers: '
                    f"[{str(src_list[edge]).strip()}] and [{str(node).strip()}] "
                    '(check for multiple assignments with "<<=" or accidental mixing '
                    'of "|=" and "<<=")'
                )
                raise PyrtlError(msg)
            src_list[edge] = node

        def add_wire_dst(edge, node):
            if edge in dst_list:
                # if node in dst_list[edge]:
                #     raise PyrtlError("The net already exists in the graph")
                dst_list[edge].append(node)
            else:
                dst_list[edge] = [node]

        if include_virtual_nodes:
            from pyrtl.wire import Const, Input, Output

            for wire in self.wirevector_subset((Input, Const)):
                add_wire_src(wire, wire)

            for wire in self.wirevector_subset(Output):
                add_wire_dst(wire, wire)

        for net in self.logic:
            for arg in set(
                net.args
            ):  # prevents unexpected duplicates when doing b <<= a & a
                add_wire_dst(arg, net)
            for dest in net.dests:
                add_wire_src(dest, net)

        return Block._NetConnectionsDict(src_list), Block._NetConnectionsDict(dst_list)

    def _repr_svg_(self):
        """IPython display support for Block."""
        from pyrtl.visualization import block_to_svg

        return block_to_svg(self)

    def __iter__(self):
        """BlockIterator iterates over the block passed on init in topographic order.
        The input is a Block, and when a LogicNet is returned it is always the case that
        all of its "parents" have already been returned earlier in the iteration.

        Note: this method will throw an error if there are loops in the logic that do
        not involve registers. Also, the order of the nets is not guaranteed to be the
        same over multiple iterations.
        """
        from pyrtl.wire import Const, Input, Register

        src_dict, dest_dict = self.net_connections()
        to_clear = self.wirevector_subset((Input, Const, Register))
        cleared = set()
        remaining = self.logic.copy()
        try:
            while len(to_clear):
                wire_to_check = to_clear.pop()
                cleared.add(wire_to_check)
                if wire_to_check in dest_dict:
                    for gate in dest_dict[
                        wire_to_check
                    ]:  # loop over logicnets not yet returned
                        if all(
                            arg in cleared for arg in gate.args
                        ):  # if all args ready
                            yield gate
                            remaining.remove(gate)
                            if gate.op != "r":
                                to_clear.update(gate.dests)
        except KeyError as e:
            msg = "Cannot Iterate through malformed block"
            raise PyrtlError(msg) from e

        if len(remaining) != 0:
            from pyrtl.helperfuncs import find_and_print_loop

            find_and_print_loop(self)
            msg = "Failure in Block Iterator due to non-register loops"
            raise PyrtlError(msg)

    def sanity_check(self):
        """Verify the ``Block``'s integrity. Raise an exception if there is an issue.

        ``sanity_check`` does not change any of the ``Block``'s state. It only verifies
        that the ``Block``'s data structures are internally consistent.

        :raise PyrtlError: If the ``Block`` is malformed.
        """
        from pyrtl.helperfuncs import get_stack, get_stacks
        from pyrtl.wire import Const, Input, Output

        # check for valid LogicNets (and wires)
        for net in self.logic:
            self.sanity_check_net(net)

        for w in self.wirevector_subset():
            if w.bitwidth is None:
                msg = (
                    f'error, missing bitwidth for WireVector "{w.name}" \n\n '
                    f"{get_stack(w)}"
                )
                raise PyrtlError(msg)

        # check for unique names
        wirevector_names_set = {x.name for x in self.wirevector_set}
        if len(self.wirevector_set) != len(wirevector_names_set):
            wirevector_names_list = [x.name for x in self.wirevector_set]
            for w in wirevector_names_set:
                wirevector_names_list.remove(w)
            msg = (
                "Duplicate wire names found for the following different signals: "
                f'{wirevector_names_list} (make sure you are not using "tmp" or '
                '"const_" as a signal name because those are reserved for internal use)'
            )
            raise PyrtlError(msg)

        # The following line also checks for duplicate wire drivers
        wire_src_dict, wire_dst_dict = self.net_connections()
        dest_set = set(wire_src_dict.keys())
        arg_set = set(wire_dst_dict.keys())
        full_set = dest_set | arg_set
        connected_minus_allwires = full_set.difference(self.wirevector_set)
        if len(connected_minus_allwires) > 0:
            bad_wire_names = "\n    ".join(str(x) for x in connected_minus_allwires)
            msg = (
                f"Unknown wires found in net:\n {bad_wire_names} \n\n "
                f"{get_stacks(*connected_minus_allwires)}"
            )
            raise PyrtlError(msg)

        all_input_and_consts = self.wirevector_subset((Input, Const))

        # Check for wires that aren't connected to anything (inputs and consts can be
        # unconnected)
        allwires_minus_connected = self.wirevector_set.difference(full_set)
        allwires_minus_connected = allwires_minus_connected.difference(
            all_input_and_consts
        )
        if len(allwires_minus_connected) > 0:
            bad_wire_names = "\n    ".join(str(x) for x in allwires_minus_connected)
            msg = (
                f"Wires declared but not connected:\n {bad_wire_names} \n\n "
                f"{get_stacks(*allwires_minus_connected)}"
            )
            raise PyrtlError(msg)

        # Check for wires that are inputs to a logicNet, but are not block inputs and
        # are never driven.
        ins = arg_set.difference(dest_set)
        undriven = ins.difference(all_input_and_consts)
        if len(undriven) > 0:
            msg = (
                f"Wires used but never driven: {[w.name for w in undriven]} \n\n "
                f"{get_stacks(*undriven)}"
            )
            raise PyrtlError(msg)

        # Check for async memories not specified as such
        self.sanity_check_memory_sync(wire_src_dict)

        # Check that all mappings in wirevector_by_name are consistent
        bad_wv_by_name = [w for n, w in self.wirevector_by_name.items() if n != w.name]
        if bad_wv_by_name:
            raise PyrtlInternalError(
                "Wires with inconsistent entry in wirevector_by_name "
                "dict: %s" % [w.name for w in bad_wv_by_name]
            )

        # Check that all wires are in wirevector_by_name
        wv_by_name_set = set(self.wirevector_by_name.keys())
        missing_wires = wirevector_names_set.difference(wv_by_name_set)
        if missing_wires:
            msg = (
                "Missing entries in wirevector_by_name for the "
                f"following wires: {missing_wires}"
            )
            raise PyrtlInternalError(msg)

        unknown_wires = wv_by_name_set.difference(wirevector_names_set)
        if unknown_wires:
            msg = f"Unknown wires found in wirevector_by_name: {unknown_wires}"
            raise PyrtlInternalError(msg)

        if debug_mode:
            # Check for wires that are destinations of a logicNet, but are not outputs
            # and are never used as args.
            outs = dest_set.difference(arg_set)
            unused = outs.difference(self.wirevector_subset(Output))
            if len(unused) > 0:
                names = [w.name for w in unused]
                print(f"Warning: Wires driven but never used {{ {names} }} ")
                print(get_stacks(*unused))

    def sanity_check_memory_sync(self, wire_src_dict=None):
        """Check that all memories are synchronous unless explicitly specified as async.

        While the semantics of 'm' memories reads is asynchronous, if you want your
        design to use a block ram (on an FPGA or otherwise) you want to make sure the
        index is available at the beginning of the clock edge. This check will walk the
        logic structure and throw an error on any memory if finds that has an index that
        is not ready at the beginning of the cycle.
        """
        sync_mems = {
            m for m in self.logic_subset("m") if not m.op_param[1].asynchronous
        }
        if not sync_mems:
            return  # nothing to check here

        if wire_src_dict is None:
            wire_src_dict, wdd = self.net_connections()

        from pyrtl.wire import Const, Input

        sync_src = "r"
        sync_prop = "wcs"
        for net in sync_mems:
            wires_to_check = list(net.args)
            while wires_to_check:
                wire = wires_to_check.pop()
                if isinstance(wire, (Input, Const)):
                    continue
                src_net = wire_src_dict[wire]
                if src_net.op == sync_src:
                    continue
                if src_net.op in sync_prop:
                    wires_to_check.extend(src_net.args)
                else:
                    msg = (
                        f'memory "{net.op_param[1].name}" is not specified as '
                        f'asynchronous but has an index "{net.args[0].name}" that is '
                        f'not ready at the start of the cycle due to net "{src_net}"'
                    )
                    raise PyrtlError(msg)

    def sanity_check_wirevector(self, w):
        """Check that w is a valid WireVector type."""
        from pyrtl.wire import WireVector

        if not isinstance(w, WireVector):
            msg = (
                f'error attempting to pass an input of type "{type(w)}" instead of '
                "WireVector"
            )
            raise PyrtlError(msg)

    def sanity_check_memblock(self, m):
        """Check that m is a valid memblock type."""
        from pyrtl.memory import MemBlock

        if not isinstance(m, MemBlock):
            msg = (
                f'error attempting to pass an input of type "{type(m)}" instead of '
                "MemBlock"
            )
            raise PyrtlError(msg)

    def sanity_check_net(self, net):
        """Check that net is a valid LogicNet."""
        from pyrtl.memory import MemBlock
        from pyrtl.wire import Const, Input, Output, Register

        # general sanity checks that apply to all operations
        if not isinstance(net, LogicNet):
            msg = "error, net must be of type LogicNet"
            raise PyrtlInternalError(msg)
        if not isinstance(net.args, tuple):
            msg = "error, LogicNet args must be tuple"
            raise PyrtlInternalError(msg)
        if not isinstance(net.dests, tuple):
            msg = "error, LogicNet dests must be tuple"
            raise PyrtlInternalError(msg)
        for w in net.args + net.dests:
            self.sanity_check_wirevector(w)
            if w._block is not self:
                msg = "error, net references different block"
                raise PyrtlInternalError(msg)
            if w not in self.wirevector_set:
                msg = f'error, net with unknown source "{w.name}"'
                raise PyrtlInternalError(msg)

        # checks that input and output WireVectors are not misused
        bad_dests = set(filter(lambda w: isinstance(w, (Input, Const)), net.dests))
        if bad_dests:
            msg = "error, Inputs, Consts cannot be destinations to a net ({})".format(
                ",".join(map(str, bad_dests))
            )
            raise PyrtlInternalError(msg)
        bad_args = set(filter(lambda w: isinstance(w, (Output)), net.args))
        if bad_args:
            msg = "error, Outputs cannot be arguments for a net ({})".format(
                ",".join(map(str, bad_args))
            )
            raise PyrtlInternalError(msg)

        if net.op not in self.legal_ops:
            msg = f'error, net op "{net.op}" not from acceptable set {self.legal_ops}'
            raise PyrtlInternalError(msg)

        # operation-specific checks on arguments
        if net.op in "w~rsm" and len(net.args) != 1:
            msg = "error, op only allowed 1 argument"
            raise PyrtlInternalError(msg)
        if net.op in "&|^n+-*<>=" and len(net.args) != 2:
            msg = "error, op only allowed 2 arguments"
            raise PyrtlInternalError(msg)
        if net.op == "x":
            if len(net.args) != 3:
                msg = "error, op only allowed 3 arguments"
                raise PyrtlInternalError(msg)
            if net.args[1].bitwidth != net.args[2].bitwidth:
                msg = "error, args have mismatched bitwidths"
                raise PyrtlInternalError(msg)
            if net.args[0].bitwidth != 1:
                msg = "error, mux select must be a single bit"
                raise PyrtlInternalError(msg)
        if net.op == "@" and len(net.args) != 3:
            msg = "error, op only allowed 3 arguments"
            raise PyrtlInternalError(msg)
        if net.op in "&|^n+-*<>=" and net.args[0].bitwidth != net.args[1].bitwidth:
            msg = "error, args have mismatched bitwidths"
            raise PyrtlInternalError(msg)
        if net.op in "m@" and net.args[0].bitwidth != net.op_param[1].addrwidth:
            msg = "error, mem addrwidth mismatch"
            raise PyrtlInternalError(msg)
        if net.op == "@" and net.args[1].bitwidth != net.op_param[1].bitwidth:
            msg = "error, mem bitwidth mismatch"
            raise PyrtlInternalError(msg)
        if net.op == "@" and net.args[2].bitwidth != 1:
            msg = "error, mem write enable must be 1 bit"
            raise PyrtlInternalError(msg)

        # operation-specific checks on op_params
        if net.op in "w~&|^n+-*<>=xcr" and net.op_param is not None:
            msg = "error, op_param should be None"
            raise PyrtlInternalError(msg)
        if net.op == "s":
            if not isinstance(net.op_param, tuple):
                msg = "error, select op requires tuple op_param"
                raise PyrtlInternalError(msg)
            for p in net.op_param:
                if not isinstance(p, int):
                    msg = "error, select op_param requires ints"
                    raise PyrtlInternalError(msg)
                if p < 0 or p >= net.args[0].bitwidth:
                    msg = "error, op_param out of bounds"
                    raise PyrtlInternalError(msg)
        if net.op in "m@":
            if not isinstance(net.op_param, tuple):
                msg = "error, mem op requires tuple op_param"
                raise PyrtlInternalError(msg)
            if len(net.op_param) != 2:
                msg = "error, mem op requires 2 op_params in tuple"
                raise PyrtlInternalError(msg)
            if not isinstance(net.op_param[0], int):
                msg = "error, mem op requires first operand as int"
                raise PyrtlInternalError(msg)
            if not isinstance(net.op_param[1], MemBlock):
                msg = "error, mem op requires second operand of a memory type"
                raise PyrtlInternalError(msg)

        # operation-specific checks on destinations
        if net.op in "w~&|^n+-*<>=xcsrm" and len(net.dests) != 1:
            msg = "error, op only allowed 1 destination"
            raise PyrtlInternalError(msg)
        if net.op == "@" and net.dests != ():
            msg = "error, mem write dest should be empty tuple"
            raise PyrtlInternalError(msg)
        if net.op == "r" and not isinstance(net.dests[0], Register):
            msg = "error, dest of next op should be a Register"
            raise PyrtlInternalError(msg)

        # check destination validity
        if net.op in "w~&|^nr" and net.dests[0].bitwidth > net.args[0].bitwidth:
            msg = "error, upper bits of destination unassigned"
            raise PyrtlInternalError(msg)
        if net.op in "<>=" and net.dests[0].bitwidth != 1:
            msg = "error, destination should be of bitwidth=1"
            raise PyrtlInternalError(msg)
        if net.op in "+-" and net.dests[0].bitwidth > net.args[0].bitwidth + 1:
            msg = "error, upper bits of destination unassigned"
            raise PyrtlInternalError(msg)
        if net.op == "*" and net.dests[0].bitwidth > 2 * net.args[0].bitwidth:
            msg = "error, upper bits of destination unassigned"
            raise PyrtlInternalError(msg)
        if net.op == "x" and net.dests[0].bitwidth > net.args[1].bitwidth:
            msg = "error, upper bits of mux output undefined"
            raise PyrtlInternalError(msg)
        if net.op == "c" and net.dests[0].bitwidth > sum(x.bitwidth for x in net.args):
            msg = "error, upper bits of concat output undefined"
            raise PyrtlInternalError(msg)
        if net.op == "s" and net.dests[0].bitwidth > len(net.op_param):
            msg = "error, upper bits of select output undefined"
            raise PyrtlInternalError(msg)
        if net.op == "m" and net.dests[0].bitwidth != net.op_param[1].bitwidth:
            msg = "error, mem read dest bitwidth mismatch"
            raise PyrtlInternalError(msg)


class PostSynthBlock(Block):
    """This is a block with extra metadata required to maintain the pre-synthesis
    interface during post-synthesis.
    """

    io_map: dict[WireVector, list[WireVector]]
    """
    A map from old IO :class:`WireVector` to a :class:`list` of new IO
    :class:`WireVectors<WireVector>` it maps to; this is a :class:`list` because for
    unmerged IO vectors, each old N-bit IO :class:`WireVector` maps to N new 1-bit IO
    :class:`WireVectors<WireVector>`.
    """
    reg_map: dict[Register, list[Register]]
    """
    A map from old :class:`Register` to a :class:`list` of new
    :class:`Registers<Register>`; a :class:`list` because post-synthesis, each N-bit
    :class:`Register` has been mapped to N 1-bit :class:`Registers<Register>`.
    """
    mem_map: dict[MemBlock, MemBlock]
    """
    A map from old :class:`MemBlock` to the new :class:`MemBlock`.
    """

    def __init__(self):
        super().__init__()
        self.io_map = collections.defaultdict(list)
        self.reg_map = collections.defaultdict(list)
        self.mem_map = {}


# -----------------------------------------------------------------------
#          __   __               __      __        __   __
#    |  | /  \ |__) |__/ | |\ | / _`    |__) |    /  \ /  ` |__/
#    |/\| \__/ |  \ |  \ | | \| \__>    |__) |___ \__/ \__, |  \
#

# Right now we use singleton_block to store the one global block, but in the future we
# should support multiple Blocks. The argument "singleton_block" should never be passed.
_singleton_block = Block()

# settings help tweak the behavior of pyrtl as needed, especially when there is a trade
# off between speed and debugability. These are useful for developers to adjust
# behaviors in the different modes but should not be set directly by users.
debug_mode = False
_setting_keep_wirevector_call_stack = False
_setting_slower_but_more_descriptive_tmps = False


def _get_debug_mode():
    return debug_mode


def _get_useful_callpoint_name():
    """Attempts to find the lowest user-level call into the PyRTL module.

    This function walks back the current frame stack attempting to find the first frame
    that is not part of the pyrtl module. The filename (stripped of path and .py
    extension) and line number of that call are returned. This point should be the point
    where the user-level code is making the call to some pyrtl intrisic (for example,
    calling "mux"). If the attempt to find the callpoint fails for any reason, None is
    returned.

    :return (string, int) or None: the file name and line number respectively
    """
    if not _setting_slower_but_more_descriptive_tmps:
        return None

    loc = None
    frame_stack = inspect.stack()
    try:
        for frame in frame_stack:
            modname = inspect.getmodule(frame[0]).__name__
            if not modname.startswith("pyrtl."):
                full_filename = frame[0].f_code.co_filename
                filename = full_filename.split("/")[-1].rstrip(".py")
                lineno = frame[0].f_lineno
                loc = (filename, lineno)
                break
    except Exception:
        loc = None
    finally:
        del frame_stack
    return loc


def working_block(block: Block = None) -> Block:
    """Convenience function for capturing the current working block.

    If a ``block`` is not passed, or if the ``block`` passed is ``None``, then this will
    return the "current working block". However, if a ``block`` is passed in it will
    simply return ``block`` instead. This feature is useful in allowing functions to
    "override" the current working block.
    """

    if block is None:
        return _singleton_block
    if not isinstance(block, Block):
        msg = "error, expected instance of Block as block argument"
        raise PyrtlError(msg)
    return block


def reset_working_block():
    """Reset the working block to be empty."""
    global _singleton_block
    _singleton_block = Block()


class set_working_block:
    """Set the working block to be the block passed as argument. Compatible with the
    ``with`` statement.

    Sanity checks will only be run if the new block is different from the original
    block.
    """

    @staticmethod
    def _set_working_block(block, no_sanity_check=False):
        global _singleton_block
        if not isinstance(block, Block):
            msg = "error, expected instance of Block as block argument"
            raise PyrtlError(msg)
        if block is not _singleton_block:  # don't update if the blocks are the same
            if not no_sanity_check:
                block.sanity_check()
            _singleton_block = block

    def __init__(self, block, no_sanity_check=False):
        self.old_block = working_block()  # for with statement compatibility
        self._set_working_block(working_block(block), no_sanity_check)

    def __enter__(self):
        return self.old_block

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._set_working_block(self.old_block, no_sanity_check=True)


def temp_working_block():
    """Set the working block to be new temporary block.

    If used with the ``with`` statement the block will be reset to the original value
    (at the time of call) at exit of the context.
    """
    return set_working_block(Block())


def set_debug_mode(debug: bool = True):
    """Set the global debug mode.

    Sets the debug mode to the specified ``debug`` value. Debug mode is off by default,
    to improve performance. When debug mode is enabled, all temporary
    :class:`WireVectors<WireVector>` will be assigned names based on the line of code on
    which they were created.

    Each :class:`WireVector` will also save a copy of its call stack when constructed.
    These call stacks can be inspected as :attr:`WireVector.init_call_stack`, and they
    will appear in :meth:`Block.sanity_check` error messages.

    :param debug: Optional boolean parameter to which the debug mode will be set.
    """
    global debug_mode
    global _setting_keep_wirevector_call_stack
    global _setting_slower_but_more_descriptive_tmps
    debug_mode = debug
    _setting_keep_wirevector_call_stack = debug
    _setting_slower_but_more_descriptive_tmps = debug


_py_regex = r"^[^\d\W]\w*\Z"


class _NameIndexer:
    """Provides internal names that are based on a prefix and an index."""

    def __init__(self, internal_prefix="_sani_temp"):
        self.internal_prefix = internal_prefix
        self.internal_index = 0

    def make_valid_string(self):
        """Build a valid string based on the prefix and internal index."""
        return self.internal_prefix + str(self.next_index())

    def next_index(self):
        index = self.internal_index
        self.internal_index += 1
        return index


class _NameSanitizer(_NameIndexer):
    """Sanitizes names so they can be used in contexts that don't allow arbitrary names.

    For example, ``a.b`` is a valid ``WireVector`` name, but ``a.b`` is not a valid name
    for a Python variable. If we want to generate Python code for ``a.b`` (like
    ``FastSimulation``), the name must be sanitized.

    Sanitization first attempts to replace non-word characters (anything that's not
    alphanumeric or an underscore) with an underscore. If that didn't work, we try
    appending a unique integer value. If that still doesn't work, we generate an
    entirely new name consisting of ``internal_prefix`` followed by a unique integer
    value.

    ``make_valid_string`` must be called once to generate the sanitized version of a
    name.

    .. doctest only::

        >>> import pyrtl

    After ``make_valid_string`` has been called, the sanitized name can be retrieved
    with ``__getitem__`` any number of times. For example::

        >>> sanitizer = pyrtl.core._NameSanitizer(pyrtl.core._py_regex)

        >>> sanitizer.make_valid_string("foo.bar")
        'foo_bar'
        >>> sanitizer["foo.bar"]
        'foo_bar'
        >>> sanitizer["foo.bar"]
        'foo_bar'

        >>> sanitizer.make_valid_string("foo_bar")
        'foo_bar0'
    """

    def __init__(
        self,
        identifier_regex_str,
        internal_prefix="_sani_temp",
        extra_checks=lambda _string: True,
    ):
        if identifier_regex_str[-1] != "$":
            identifier_regex_str += "$"
        self.identifier = re.compile(identifier_regex_str)
        # Map from un-sanitized name to sanitized name.
        self.val_map = {}
        # Set of all generated sanitized names.
        self.sanitized_names = set()
        self.extra_checks = extra_checks
        super().__init__(internal_prefix)

    def __getitem__(self, name: str) -> str:
        """Return the sanitized name for an un-sanitized name that was generated by
        ``make_valid_string``.
        """
        return self.val_map[name]

    def is_valid_str(self, string: str) -> bool:
        """Return ``True`` iff ``string`` matches ``identifier_regex_str`` and satisfies
        ``extra_checks``.
        """
        return self.identifier.match(string) and self.extra_checks(string)

    def make_valid_string(self, string: str = "") -> str:
        """Generate a sanitized name from an un-sanitized name."""
        if string in self.val_map:
            msg = f"Value {string} has already been given to the sanitizer"
            raise IndexError(msg)

        def is_usable(name: str) -> bool:
            """Return ``True`` iff ``name`` can be used as a sanitized name.

            A sanitized name is usable if it ``is_valid_str``, and isn't already in use.
            """
            return self.is_valid_str(name) and name not in self.sanitized_names

        internal_name = string
        if not is_usable(internal_name):
            # Try replacing non-word characters with ``_``.
            internal_name = re.sub(r"\W", "_", string)

            if not is_usable(internal_name):
                # If that didn't work, try appending the next ``internal_index``.
                internal_name = f"{internal_name}{self.next_index()}"

                if not is_usable(internal_name):
                    # If that didn't work, generate an entirely new name starting with
                    # ``internal_prefix``.
                    internal_name = super().make_valid_string()

                    if not is_usable(internal_name):
                        msg = f"Could not generate a usable sanitized name for {string}"
                        raise PyrtlError(msg)

        self.val_map[string] = internal_name
        self.sanitized_names.add(internal_name)
        return internal_name


class _PythonSanitizer(_NameSanitizer):
    """Name Sanitizer specifically built for Python identifers."""

    def __init__(self, internal_prefix="_sani_temp"):
        super().__init__(_py_regex, internal_prefix)
        self.extra_checks = lambda s: not keyword.iskeyword(s)



import unittest

import pyrtl
import pyrtl.corecircuits
from pyrtl.rtllib import aes, testingutils


class TestAESDecrypt(unittest.TestCase):
    """
    Test vectors are retrieved from:
    http://csrc.nist.gov/publications/fips/fips197/fips-197.pdf
    """

    def setUp(self):
        pyrtl.reset_working_block()
        self.aes_decrypt = aes.AES()
        self.in_vector = pyrtl.Input(bitwidth=128, name="in_vector")
        self.out_vector = pyrtl.Output(bitwidth=128, name="out_vector")

    def test_inv_shift_rows(self):
        self.out_vector <<= self.aes_decrypt._inv_shift_rows(self.in_vector)

        in_vals = [
            0x3E1C22C0B6FCBF768DA85067F6170495,
            0x2D6D7EF03F33E334093602DD5BFB12C7,
        ]
        true_result = [
            0x3E175076B61C04678DFC2295F6A8BFC0,
            0x2DFB02343F6D12DD09337EC75B36E3F0,
        ]
        calculated_result = testingutils.sim_and_ret_out(
            self.out_vector, (self.in_vector,), (in_vals,)
        )
        self.assertEqual(calculated_result, true_result)

    def test_inv_sub_bytes(self):
        self.out_vector <<= self.aes_decrypt._sub_bytes(self.in_vector, True)

        in_vals = [
            0x3E175076B61C04678DFC2295F6A8BFC0,
            0x2DFB02343F6D12DD09337EC75B36E3F0,
        ]
        true_result = [
            0xD1876C0F79C4300AB45594ADD66FF41F,
            0xFA636A2825B339C940668A3157244D17,
        ]
        calculated_result = testingutils.sim_and_ret_out(
            self.out_vector, (self.in_vector,), (in_vals,)
        )
        self.assertEqual(calculated_result, true_result)

    def test_inv_mix_columns(self):
        self.out_vector <<= self.aes_decrypt._mix_columns(self.in_vector, True)

        in_vals = [
            0xE9F74EEC023020F61BF2CCF2353C21C7,
            0xBAA03DE7A1F9B56ED5512CBA5F414D23,
        ]
        real_res = [
            0x54D990A16BA09AB596BBF40EA111702F,
            0x3E1C22C0B6FCBF768DA85067F6170495,
        ]
        calculated_result = testingutils.sim_and_ret_out(
            self.out_vector, (self.in_vector,), (in_vals,)
        )
        self.assertEqual(calculated_result, real_res)

    def test_aes_full(self):
        aes_key = pyrtl.Input(bitwidth=128, name="aes_key")
        self.out_vector <<= self.aes_decrypt.decryption(self.in_vector, aes_key)

        ciphers = [
            0x3AD77BB40D7A3660A89ECAF32466EF97,
            0x66E94BD4EF8A2C3B884CFA59CA342B2E,
        ]
        keys = [0x2B7E151628AED2A6ABF7158809CF4F3C, 0x0]
        plain_text = [0x6BC1BEE22E409F96E93D7E117393172A, 0x0]
        calculated_result = testingutils.sim_and_ret_out(
            self.out_vector, (self.in_vector, aes_key), (ciphers, keys)
        )
        self.assertEqual(calculated_result, plain_text)

    def test_aes_state_machine(self):
        # self.longMessage = True

        aes_key = pyrtl.Input(bitwidth=128, name="aes_key")
        reset = pyrtl.Input(1)
        ready = pyrtl.Output(1, name="ready")

        decrypt_ready, decrypt_out = self.aes_decrypt.decryption_statem(
            self.in_vector, aes_key, reset
        )
        self.out_vector <<= decrypt_out
        ready <<= decrypt_ready

        sim = pyrtl.Simulation()

        sim.step(
            {
                self.in_vector: 0x69C4E0D86A7B0430D8CDB78070B4C55A,
                aes_key: 0x000102030405060708090A0B0C0D0E0F,
                reset: 1,
            }
        )

        true_vals = [
            0x69C4E0D86A7B0430D8CDB78070B4C55A,
            0x7AD5FDA789EF4E272BCA100B3D9FF59F,
            0x54D990A16BA09AB596BBF40EA111702F,
            0x3E1C22C0B6FCBF768DA85067F6170495,
            0xB458124C68B68A014B99F82E5F15554C,
            0xE8DAB6901477D4653FF7F5E2E747DD4F,
            0x36339D50F9B539269F2C092DC4406D23,
            0x2D6D7EF03F33E334093602DD5BFB12C7,
            0x3BD92268FC74FB735767CBE0C0590E2D,
            0xA7BE1A6997AD739BD8C9CA451F618B61,
            0x6353E08C0960E104CD70B751BACAD0E7,
            0x00112233445566778899AABBCCDDEEFF,
            0x00112233445566778899AABBCCDDEEFF,
        ]

        for cycle in range(1, 13):  # Bogus data for while the state machine churns
            sim.step({self.in_vector: 0x0, aes_key: 0x1, reset: 0})
            circuit_out = sim.tracer.trace["out_vector"][cycle]
            self.assertEqual(
                circuit_out,
                true_vals[cycle],
                f"\nAssertion failed on cycle: {cycle} Gotten value: "
                f"{hex(circuit_out)}",
            )

        for ready_signal in sim.tracer.trace["ready"][:11]:
            self.assertEqual(ready_signal, 0)

        for ready_signal in sim.tracer.trace["ready"][11:]:
            self.assertEqual(ready_signal, 1)


class TestAESEncrypt(unittest.TestCase):
    """
    Test vectors are retrieved from:
    http://csrc.nist.gov/publications/fips/fips197/fips-197.pdf
    """

    def setUp(self):
        pyrtl.reset_working_block()
        self.aes_encrypt = aes.AES()
        self.in_vector = pyrtl.Input(bitwidth=128, name="in_vector")
        self.out_vector = pyrtl.Output(bitwidth=128, name="out_vector")

    def test_shift_rows(self):
        self.out_vector <<= self.aes_encrypt._shift_rows(self.in_vector)

        in_vals = [
            0x3B59CB73FCD90EE05774222DC067FB68,
            0xB415F8016858552E4BB6124C5F998A4C,
        ]
        true_result = [
            0x3BD92268FC74FB735767CBE0C0590E2D,
            0xB458124C68B68A014B99F82E5F15554C,
        ]
        calculated_result = testingutils.sim_and_ret_out(
            self.out_vector, (self.in_vector,), (in_vals,)
        )
        self.assertEqual(calculated_result, true_result)

    def test_sub_bytes(self):
        self.out_vector <<= self.aes_encrypt._sub_bytes(self.in_vector)

        in_vals = [
            0x4915598F55E5D7A0DACA94FA1F0A63F7,
            0xC62FE109F75EEDC3CC79395D84F9CF5D,
        ]
        true_result = [
            0x3B59CB73FCD90EE05774222DC067FB68,
            0xB415F8016858552E4BB6124C5F998A4C,
        ]
        calculated_result = testingutils.sim_and_ret_out(
            self.out_vector, (self.in_vector,), (in_vals,)
        )
        self.assertEqual(calculated_result, true_result)

    def test_mix_columns(self):
        self.out_vector <<= self.aes_encrypt._mix_columns(self.in_vector)

        in_vals = [
            0x6353E08C0960E104CD70B751BACAD0E7,
            0xA7BE1A6997AD739BD8C9CA451F618B61,
        ]
        real_res = [
            0x5F72641557F5BC92F7BE3B291DB9F91A,
            0xFF87968431D86A51645151FA773AD009,
        ]
        calculated_result = testingutils.sim_and_ret_out(
            self.out_vector, (self.in_vector,), (in_vals,)
        )
        self.assertEqual(calculated_result, real_res)

    def test_aes_full(self):
        aes_key = pyrtl.Input(bitwidth=128, name="aes_key")
        self.out_vector <<= self.aes_encrypt.encryption(self.in_vector, aes_key)

        plain_text = [0x00112233445566778899AABBCCDDEEFF, 0x0]
        keys = [0x000102030405060708090A0B0C0D0E0F, 0x0]
        ciphers = [
            0x69C4E0D86A7B0430D8CDB78070B4C55A,
            0x66E94BD4EF8A2C3B884CFA59CA342B2E,
        ]
        calculated_result = testingutils.sim_and_ret_out(
            self.out_vector, (self.in_vector, aes_key), (plain_text, keys)
        )
        self.assertEqual(calculated_result, ciphers)

    def test_aes_state_machine(self):
        # self.longMessage = True

        aes_key = pyrtl.Input(bitwidth=128, name="aes_key")
        reset = pyrtl.Input(1)
        ready = pyrtl.Output(1, name="ready")

        encrypt_ready, encrypt_out = self.aes_encrypt.encrypt_state_m(
            self.in_vector, aes_key, reset
        )
        self.out_vector <<= encrypt_out
        ready <<= encrypt_ready

        sim = pyrtl.Simulation()

        sim.step(
            {
                self.in_vector: 0x00112233445566778899AABBCCDDEEFF,
                aes_key: 0x000102030405060708090A0B0C0D0E0F,
                reset: 1,
            }
        )

        true_vals = [
            0x00112233445566778899AABBCCDDEEFF,
            0x00102030405060708090A0B0C0D0E0F0,
            0x89D810E8855ACE682D1843D8CB128FE4,
            0x4915598F55E5D7A0DACA94FA1F0A63F7,
            0xFA636A2825B339C940668A3157244D17,
            0x247240236966B3FA6ED2753288425B6C,
            0xC81677BC9B7AC93B25027992B0261996,
            0xC62FE109F75EEDC3CC79395D84F9CF5D,
            0xD1876C0F79C4300AB45594ADD66FF41F,
            0xFDE3BAD205E5D0D73547964EF1FE37F1,
            0xBD6E7C3DF2B5779E0B61216E8B10B689,
            0x69C4E0D86A7B0430D8CDB78070B4C55A,
            0x69C4E0D86A7B0430D8CDB78070B4C55A,
        ]

        for cycle in range(1, 13):  # Bogus data for while the state machine churns
            sim.step({self.in_vector: 0x0, aes_key: 0x1, reset: 0})
            circuit_out = sim.tracer.trace["out_vector"][cycle]
            # sim.tracer.render_trace(symbol_len=40)
            self.assertEqual(
                circuit_out,
                true_vals[cycle],
                f"\nAssertion failed on cycle: {cycle} Gotten value: "
                f"{hex(circuit_out)}",
            )

        for ready_signal in sim.tracer.trace["ready"][:11]:
            self.assertEqual(ready_signal, 0)

        for ready_signal in sim.tracer.trace["ready"][11:]:
            self.assertEqual(ready_signal, 1)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import _ctypes
import ctypes
import platform
import shutil
import subprocess
import sys
import tempfile
import warnings
from collections.abc import Mapping
from os import path

from pyrtl.core import Block, working_block
from pyrtl.helperfuncs import infer_val_and_bitwidth
from pyrtl.memory import MemBlock, RomBlock
from pyrtl.pyrtlexceptions import PyrtlError, PyrtlInternalError
from pyrtl.simulation import SimulationTrace, _trace_sort_key
from pyrtl.wire import Const, Input, Output, Register, WireVector

__all__ = ["CompiledSimulation"]


class DllMemInspector(Mapping):
    """Dictionary-like access to a hashmap in a CompiledSimulation."""

    def __init__(self, sim, mem):
        self._aw = mem.addrwidth
        self._limbs = sim._limbs(mem)
        self._vn = vn = sim.varname[mem]
        self._mem = ctypes.c_void_p.in_dll(sim._dll, vn)
        self._sim = sim  # keep reference to avoid freeing dll

    def __getitem__(self, ind):
        arr = self._sim._mem_lookup(self._mem, ind)
        val = 0
        for n in reversed(range(self._limbs)):
            val <<= 64
            val |= arr[n]
        return val

    def __iter__(self):
        return iter(range(len(self)))

    def __len__(self):
        return 1 << self._aw

    def __eq__(self, other):
        if (
            isinstance(other, DllMemInspector)
            and self._sim is other._sim
            and self._vn == other._vn
        ):
            return True
        return all(self[x] == other.get(x, 0) for x in self)

    def __hash__(self):
        return hash(self._sim) ^ hash(self._vn)


class CompiledSimulation:
    """Simulate a block by generating, compiling, and running C code.

    ``CompiledSimulation`` provides significant execution speed improvements over
    :class:`FastSimulation`, at the cost of even longer start-up time. Generally this
    will do better than :class:`FastSimulation` for simulations requiring over 1000
    steps.

    ``CompiledSimulation`` is not built to be a debugging tool, though it may help with
    debugging. Note that only :class:`Input` and :class:`Output` wires can be traced
    with ``CompiledSimulation``.

    .. note::

        For very large circuits, :class:`FastSimulation` can sometimes be a better
        choice than ``CompiledSimulation`` because ``CompiledSimulation`` will generate
        an extremely large ``.c`` file, which can take prohibitively long to compile and
        optimize. :class:`FastSimulation` will generate an extremely large ``.py``
        file, but Python will interpret that generated code as needed, instead of trying
        to process all the generated code at once.

    .. WARNING::

        This code is still experimental, but has been used on designs of significant
        scale to good effect.

    To use ``CompiledSimulation``, you'll need:

    - A 64-bit processor

    - GCC (tested on version 4.8.4)

    - A 64-bit build of Python

    If using the multiplication operand, only some architectures are supported:

    - ``x86-64`` / ``amd64``

    - ``arm64`` / ``aarch64``

    - ``mips64`` (untested)

    ``default_value`` is currently only implemented for :class:`Registers<Register>`,
    not :class:`MemBlocks<MemBlock>`.

    ``CompiledSimulation`` is a drop-in replacement for :class:`Simulation`, so the two
    classes share the same interface. See :class:`Simulation` for interface
    documentation, and more details about PyRTL simulations.
    """

    def __init__(
        self,
        tracer: SimulationTrace = True,
        register_value_map: dict[Register, int] | None = None,
        memory_value_map: dict[MemBlock, dict[int, int]] | None = None,
        default_value: int = 0,
        block: Block = None,
    ):
        if memory_value_map is None:
            memory_value_map = {}
        if register_value_map is None:
            register_value_map = {}
        self._dll = self._dir = None
        self.block = working_block(block)
        self.block.sanity_check()

        if tracer is True:
            tracer = SimulationTrace()
        self.tracer = tracer
        self._remove_untraceable()

        self.default_value = default_value
        self._regmap = {}  # Updated below
        self._memmap = memory_value_map
        self._uid_counter = 0
        self.varname = {}  # mapping from wires and memories to C variables

        for r in self.block.wirevector_subset(Register):
            rval = register_value_map.get(r, r.reset_value)
            if rval is None:
                rval = self.default_value
            self._regmap[r] = rval

        self.tracer._set_initial_values(
            default_value, register_value_map, memory_value_map
        )

        self._create_dll()
        self._initialize_mems()

    def inspect_mem(self, mem: MemBlock) -> dict[int, int]:
        return DllMemInspector(self, mem)

    def inspect(self, w: str) -> int:
        if isinstance(w, WireVector):
            w = w.name
        try:
            vals = self.tracer.trace[w]
        except KeyError:
            pass
        else:
            if not vals:
                msg = "No context available. Please run a simulation step"
                raise PyrtlError(msg)
            return vals[-1]
        msg = "CompiledSimulation does not support inspecting internal WireVectors"
        raise PyrtlError(msg)

    def step(self, provided_inputs: dict[str, int] | None = None, inputs=None):
        if provided_inputs is None:
            provided_inputs = {}
        if inputs is not None:
            warnings.warn(
                "CompiledSimulation.step: `inputs` was renamed to `provided_inputs`",
                DeprecationWarning,
                stacklevel=2,
            )
            provided_inputs = inputs
        self.run([provided_inputs])

    def step_multiple(
        self,
        provided_inputs: dict[str, list[int]] | None = None,
        expected_outputs: dict[str, int] | None = None,
        nsteps: int | None = None,
        file=sys.stdout,
        stop_after_first_error: bool = False,
    ):
        if expected_outputs is None:
            expected_outputs = {}
        if provided_inputs is None:
            provided_inputs = {}
        if not nsteps and len(provided_inputs) == 0:
            msg = "need to supply either input values or a number of steps to simulate"
            raise PyrtlError(msg)

        if len(provided_inputs) > 0:
            longest = sorted(
                provided_inputs.items(), key=lambda t: len(t[1]), reverse=True
            )[0]
            msteps = len(longest[1])
            if nsteps:
                if nsteps > msteps:
                    msg = (
                        "nsteps is specified but is greater than the number of values "
                        "supplied for each input"
                    )
                    raise PyrtlError(msg)
            else:
                nsteps = msteps

        if nsteps < 1:
            msg = "must simulate at least one step"
            raise PyrtlError(msg)

        if list(filter(lambda value: len(value) < nsteps, provided_inputs.values())):
            msg = (
                "must supply a value for each provided wire for each step of simulation"
            )
            raise PyrtlError(msg)

        if list(filter(lambda value: len(value) < nsteps, expected_outputs.values())):
            msg = (
                "any expected outputs must have a supplied value each step of "
                "simulation"
            )
            raise PyrtlError(msg)

        failed = []
        for i in range(nsteps):
            self.step({w: int(v[i]) for w, v in provided_inputs.items()})

            for expvar in expected_outputs:
                expected = expected_outputs[expvar][i]
                if expected == "?":
                    continue
                expected = int(expected)
                actual = self.inspect(expvar)
                if expected != actual:
                    failed.append((i, expvar, expected, actual))

            if failed and stop_after_first_error:
                break

        if failed:
            if stop_after_first_error:
                s = "(stopped after step with first error):"
            else:
                s = "on one or more steps:"
            file.write("Unexpected output " + s + "\n")
            file.write(
                "{:>5} {:>10} {:>8} {:>8}\n".format(
                    "step", "name", "expected", "actual"
                )
            )

            def _sort_tuple(t):
                # Sort by step and then wire name
                return (t[0], _trace_sort_key(t[1]))

            failed_sorted = sorted(failed, key=_sort_tuple)
            for step, name, expected, actual in failed_sorted:
                file.write(f"{step:>5} {name:>10} {expected:>8} {actual:>8}\n")
            file.flush()

    def run(self, inputs: list[dict[str, int]]):
        """Run many steps of the ``CompiledSimulation``.

        :meth:`CompiledSimulation.step` and :meth:`CompiledSimulation.step_multiple` are
        wrappers around this lower-level method.

        :param inputs: A list of input mappings for each step; its length is the number
            of steps to be executed.
        """
        steps = len(inputs)
        # create i/o arrays of the appropriate length
        ibuf_type = ctypes.c_uint64 * (steps * self._ibufsz)
        obuf_type = ctypes.c_uint64 * (steps * self._obufsz)
        ibuf = ibuf_type()
        obuf = obuf_type()
        # these array will be passed to _crun
        self._crun.argtypes = [ctypes.c_uint64, ibuf_type, obuf_type]

        # build the input array
        for n, inmap in enumerate(inputs):
            for w in inmap:
                if isinstance(w, WireVector):
                    name = w.name
                else:
                    name = w
                start, count = self._inputpos[name]
                start += n * self._ibufsz
                val = inmap[w]
                val = infer_val_and_bitwidth(val, bitwidth=self._inputbw[name]).value
                # pack input
                for pos in range(start, start + count):
                    ibuf[pos] = val & ((1 << 64) - 1)
                    val >>= 64

        # run the simulation
        self._crun(steps, ibuf, obuf)

        # save traced wires
        for name in self.tracer.trace:
            rname = self._probe_mapping.get(name, name)
            if rname in self._outputpos:
                start, count = self._outputpos[rname]
                buf, sz = obuf, self._obufsz
            elif rname in self._inputpos:
                start, count = self._inputpos[rname]
                buf, sz = ibuf, self._ibufsz
            else:
                msg = "Untraceable wire in tracer"
                raise PyrtlInternalError(msg)
            res = []
            for _step in range(steps):
                val = 0
                # unpack output
                for pos in reversed(range(start, start + count)):
                    val <<= 64
                    val |= buf[pos]
                res.append(val)
                start += sz
            self.tracer.trace[name].extend(res)

    def _traceable(self, wv):
        """Check if wv is able to be traced.

        If it is traceable due to a probe, record that probe in _probe_mapping.
        """
        if isinstance(wv, (Input, Output)):
            return True
        for net in self.block.logic:
            if (
                net.op == "w"
                and net.args[0].name == wv.name
                and isinstance(net.dests[0], Output)
            ):
                self._probe_mapping[wv.name] = net.dests[0].name
                return True
        return False

    def _remove_untraceable(self):
        """Remove from the tracer those wires that CompiledSimulation cannot track.

        Create _probe_mapping for wires only traceable via probes.
        """
        self._probe_mapping = {}
        wvs = {wv for wv in self.tracer.wires_to_track if self._traceable(wv)}
        self.tracer.wires_to_track = wvs
        self.tracer._wires = {wv.name: wv for wv in wvs}
        self.tracer.trace.__init__(wvs)

    def _create_dll(self):
        """Create a dynamically-linked library implementing the simulation logic."""
        self._dir = tempfile.mkdtemp()
        with open(path.join(self._dir, "pyrtlsim.c"), "w") as f:
            self._create_code(lambda s: f.write(s + "\n"))
        if platform.system() == "Darwin":
            shared = "-dynamiclib"
            march = ""
        else:
            shared = "-shared"
            march = "-march=native"

        subprocess.check_call(
            [
                "gcc",
                "-O0",
                march,
                "-std=c99",
                "-m64",
                shared,
                "-fPIC",
                path.join(self._dir, "pyrtlsim.c"),
                "-o",
                path.join(self._dir, "pyrtlsim.so"),
            ],
            shell=(platform.system() == "Windows"),
        )
        self._dll = ctypes.CDLL(path.join(self._dir, "pyrtlsim.so"))
        self._crun = self._dll.sim_run_all
        self._crun.restype = None  # argtypes set on use
        self._initialize_mems = self._dll.initialize_mems
        self._initialize_mems.restype = None
        self._mem_lookup = self._dll.lookup
        self._mem_lookup.restype = ctypes.POINTER(ctypes.c_uint64)

    def _limbs(self, w):
        """Number of 64-bit words needed to store value of wire."""
        return (w.bitwidth + 63) // 64

    def _makeini(self, w, v):
        """C initializer string for a wire with a given value."""
        pieces = []
        for _ in range(self._limbs(w)):
            pieces.append(hex(v & ((1 << 64) - 1)))
            v >>= 64
        return ",".join(pieces).join("{}")

    def _romwidth(self, m):
        """Bitwidth of integer type sufficient to hold rom entry.

        On large memories, returns 64; an array will be needed.
        """
        if m.bitwidth <= 8:
            return 8
        if m.bitwidth <= 16:
            return 16
        if m.bitwidth <= 32:
            return 32
        return 64

    def _makemask(self, dest, res, pos):
        """Create a bitmask.

        The value being masked is of width `res`. Limb number `pos` of `dest` is being
        assigned to.
        """
        if (res is None or dest.bitwidth < res) and 0 < (dest.bitwidth - 64 * pos) < 64:
            return f"&0x{(1 << (dest.bitwidth % 64)) - 1:X}"
        return ""

    def _getarglimb(self, arg, n):
        """Get the nth limb of the given wire.

        Returns '0' when the wire does not have sufficient limbs.
        """
        return f"{self.varname[arg]}[{n}]" if arg.bitwidth > 64 * n else "0"

    def _clean_name(self, prefix, obj):
        """Create a C variable name with the given prefix based on the name of obj."""
        return "{}{}_{}".format(
            prefix, self._uid(), "".join(c for c in obj.name if c.isalnum())
        )

    def _uid(self):
        """Get an auto-incrementing number suitable for use as a unique identifier."""
        x = self._uid_counter
        self._uid_counter += 1
        return x

    def _declare_roms(self, write, roms):
        for mem in roms:
            self.varname[mem] = vn = self._clean_name("m", mem)
            # extract data from mem
            romval = [mem._get_read_data(n) for n in range(1 << mem.addrwidth)]
            write(
                f"static const uint{self._romwidth(mem)}_t {vn}[]"
                f"[{self._limbs(mem)}] = {{"
            )
            for rv in romval:
                write(self._makeini(mem, rv) + ",")
            write("};")

    def _declare_mems(self, write, mems):
        for mem in mems:
            self.varname[mem] = vn = self._clean_name("m", mem)
            write("EXPORT")
            write(f"hashmap_t *{vn};")

        next_tmp = 0
        write("EXPORT")
        write("void initialize_mems() {")
        for mem in mems:
            # Create hashmap
            write(f"{self.varname[mem]} = create_hash_map(256, {self._limbs(mem)});")
            if mem in self._memmap:
                # Insert default values
                for k, v in self._memmap[mem].items():
                    write(f"val_t t{next_tmp}[] = {self._makeini(mem, v)};")
                    write(f"insert({self.varname[mem]}, {k}, t{next_tmp});")
                    next_tmp += 1
        write("}")

    def _declare_wv(self, write, w):
        self.varname[w] = vn = self._clean_name("w", w)
        if isinstance(w, Const):
            write(f"const uint64_t {vn}[{self._limbs(w)}] = {self._makeini(w, w.val)};")
        elif isinstance(w, Register):
            rval = self._regmap.get(w, w.reset_value)
            if rval is None:
                rval = self.default_value
            write(f"static uint64_t {vn}[{self._limbs(w)}] = {self._makeini(w, rval)};")
        else:
            write(f"uint64_t {vn}[{self._limbs(w)}];")

    def _build_memread(self, write, _op, param, args, dest):
        mem = param[1]
        for n in range(self._limbs(dest)):
            if isinstance(mem, RomBlock):
                write(
                    f"{self.varname[dest]}[{n}] = {self.varname[mem]}["
                    f"{self.varname[args[0]]}[0]][{n}]"
                    f"{self._makemask(dest, mem.bitwidth, n)};"
                )
            else:
                write(
                    f"{self.varname[dest]}[{n}] = lookup({self.varname[mem]}, "
                    f"{self.varname[args[0]]}[0])[{n}]"
                    f"{self._makemask(dest, mem.bitwidth, n)};"
                )

    def _build_wire(self, write, _op, _param, args, dest):
        for n in range(self._limbs(dest)):
            write(
                f"{self.varname[dest]}[{n}] = "
                f"{self.varname[args[0]]}[{n}]"
                f"{self._makemask(dest, args[0].bitwidth, n)};"
            )

    def _build_not(self, write, _op, _param, args, dest):
        for n in range(self._limbs(dest)):
            write(
                f"{self.varname[dest]}[{n}] = "
                f"(~{self.varname[args[0]]}[{n}]){self._makemask(dest, None, n)};"
            )

    def _build_bitwise(self, write, op, _param, args, dest):  # &, |, ^ only
        for n in range(self._limbs(dest)):
            arg0 = self._getarglimb(args[0], n)
            arg1 = self._getarglimb(args[1], n)
            write(
                "{dest}[{n}] = ({arg0}{op}{arg1}){mask};".format(
                    dest=self.varname[dest],
                    n=n,
                    arg0=arg0,
                    arg1=arg1,
                    op=op,
                    mask=self._makemask(
                        dest, max(args[0].bitwidth, args[1].bitwidth), n
                    ),
                )
            )

    def _build_nand(self, write, _op, _param, args, dest):
        for n in range(self._limbs(dest)):
            arg0 = self._getarglimb(args[0], n)
            arg1 = self._getarglimb(args[1], n)
            write(
                f"{self.varname[dest]}[{n}] = "
                f"(~({arg0}&{arg1})){self._makemask(dest, None, n)};"
            )

    def _build_eq(self, write, _op, _param, args, dest):
        cond = []
        for n in range(max(self._limbs(args[0]), self._limbs(args[1]))):
            arg0 = self._getarglimb(args[0], n)
            arg1 = self._getarglimb(args[1], n)
            cond.append(f"({arg0}=={arg1})")
        write(
            "{dest}[0] = {cond};".format(dest=self.varname[dest], cond="&&".join(cond))
        )

    def _build_cmp(self, write, op, _param, args, dest):  # <, > only
        cond = None
        for n in range(max(self._limbs(args[0]), self._limbs(args[1]))):
            arg0 = self._getarglimb(args[0], n)
            arg1 = self._getarglimb(args[1], n)
            c = f"({arg0}{op}{arg1})"
            if cond is None:
                cond = c
            else:
                cond = f"({c}||(({arg0}=={arg1})&&{cond}))"
        write(f"{self.varname[dest]}[0] = {cond};")

    def _build_mux(self, write, _op, _param, args, dest):
        write(f"if ({self.varname[args[0]]}[0]) {{")
        for n in range(self._limbs(dest)):
            write(
                f"{self.varname[dest]}[{n}] = "
                f"{self.varname[args[2]]}[{n}]"
                f"{self._makemask(dest, args[2].bitwidth, n)};"
            )
        write("} else {")
        for n in range(self._limbs(dest)):
            write(
                f"{self.varname[dest]}[{n}] = "
                f"{self.varname[args[1]]}[{n}]"
                f"{self._makemask(dest, args[1].bitwidth, n)};"
            )
        write("}")

    def _build_add(self, write, _op, _param, args, dest):
        write("carry = 0;")
        for n in range(self._limbs(dest)):
            arg0 = self._getarglimb(args[0], n)
            arg1 = self._getarglimb(args[1], n)
            write(f"tmp = {arg0}+{arg1};")
            write(
                "{dest}[{n}] = (tmp + carry){mask};".format(
                    dest=self.varname[dest],
                    n=n,
                    mask=self._makemask(
                        dest, max(args[0].bitwidth, args[1].bitwidth) + 1, n
                    ),
                )
            )
            write(f"carry = (tmp < {arg0})|({self.varname[dest]}[{n}] < tmp);")

    def _build_sub(self, write, _op, _param, args, dest):
        write("carry = 0;")
        for n in range(self._limbs(dest)):
            arg0 = self._getarglimb(args[0], n)
            arg1 = self._getarglimb(args[1], n)
            write(f"tmp = {arg0}-{arg1};")
            write(
                f"{self.varname[dest]}[{n}] = (tmp - carry)"
                f"{self._makemask(dest, None, n)};"
            )
            write(f"carry = (tmp > {arg0})|({self.varname[dest]}[{n}] > tmp);")

    def _build_mul(self, write, _op, _param, args, dest):
        for n in range(self._limbs(dest)):
            write(f"{self.varname[dest]}[{n}] = 0;")
        for p0 in range(self._limbs(args[0])):
            write("carry = 0;")
            arg0 = self._getarglimb(args[0], p0)
            for p1 in range(self._limbs(args[1])):
                if self._limbs(dest) <= p0 + p1:
                    break
                arg1 = self._getarglimb(args[1], p1)
                write(f"mul128({arg0}, {arg1}, tmplo, tmphi);")
                write(f"tmp = {self.varname[dest]}[{p0 + p1}];")
                write("tmplo += carry; carry = tmplo < carry; tmplo += tmp;")
                write("tmphi += carry + (tmplo < tmp); carry = tmphi;")
                write(
                    "{dest}[{p}] = tmplo{mask};".format(
                        dest=self.varname[dest],
                        p=p0 + p1,
                        mask=self._makemask(
                            dest, args[0].bitwidth + args[1].bitwidth, p0 + p1
                        ),
                    )
                )
            if self._limbs(dest) > p0 + self._limbs(args[1]):
                write(
                    "{dest}[{p}] = carry{mask};".format(
                        dest=self.varname[dest],
                        p=p0 + self._limbs(args[1]),
                        mask=self._makemask(
                            dest,
                            args[0].bitwidth + args[1].bitwidth,
                            p0 + self._limbs(args[1]),
                        ),
                    )
                )

    def _build_concat(self, write, _op, _param, args, dest):
        cattotal = sum(x.bitwidth for x in args)
        pieces = (
            (self.varname[a], lx, 0, min(64, a.bitwidth - 64 * lx))
            for a in reversed(args)
            for lx in range(self._limbs(a))
        )
        curr = next(pieces)
        for n in range(self._limbs(dest)):
            res = []
            dpos = 0
            while True:
                arg, alimb, astart, asize = curr
                res.append(f"(({arg}[{alimb}]>>{astart})<<{dpos})")
                dpos += asize
                if dpos >= dest.bitwidth - 64 * n:
                    break
                if dpos > 64:
                    curr = (arg, alimb, 64 - (dpos - asize), dpos - 64)
                    break
                curr = next(pieces)
                if dpos == 64:
                    break
            write(
                "{dest}[{n}] = ({res}){mask};".format(
                    dest=self.varname[dest],
                    n=n,
                    res="|".join(res),
                    mask=self._makemask(dest, cattotal, n),
                )
            )

    def _build_select(self, write, _op, param, args, dest):
        for n in range(self._limbs(dest)):
            bits = [
                f"((1&({self.varname[args[0]]}[{b // 64}]>>{b % 64}))<<{en})"
                for en, b in enumerate(param[64 * n : min(dest.bitwidth, 64 * (n + 1))])
            ]
            write(
                "{dest}[{n}] = {bits};".format(
                    dest=self.varname[dest], n=n, bits="|".join(bits)
                )
            )

    def _declare_mem_helpers(self, write):
        helpers = """
            typedef uint64_t val_t;

            typedef struct node
            {
                uint64_t key;
                val_t *val;
                struct node *next;
            } node_t;

            typedef struct hashmap
            {
                int size;
                int val_limbs;
                val_t *default_value;
                node_t **list;
            } hashmap_t;

            hashmap_t *create_hash_map(int size, int val_limbs)
            {
                int i;
                hashmap_t *h = (hashmap_t *) malloc(sizeof(hashmap_t));
                h->size = size;
                h->val_limbs = val_limbs;
                h->list = (node_t **) malloc(sizeof(node_t *) * size);
                h->default_value = (val_t *) malloc(sizeof(val_t) * val_limbs);
                for (i = 0; i < val_limbs; i++)
                    h->default_value[i] = 0;
                for (i = 0; i < size; i++)
                    h->list[i] = NULL;
                return h;
            }

            int hash_code(hashmap_t *h, uint64_t key)
            {
                return key % h->size;
            }

            void insert(hashmap_t *h, uint64_t key, val_t val[])
            {
                int pos = hash_code(h, key);
                struct node *list = h->list[pos];
                struct node *new_node = (node_t *) malloc(sizeof(node_t));
                struct node *temp = list;
                while (temp)
                {
                    if (temp->key == key)
                    {
                        memcpy(temp->val, val, sizeof(val_t) * h->val_limbs);
                        return;
                    }
                    temp = temp->next;
                }
                new_node->key = key;
                new_node->val = (val_t *) malloc(sizeof(val_t) * h->val_limbs);
                memcpy(new_node->val, val, sizeof(val_t) * h->val_limbs);
                new_node->next = list;
                h->list[pos] = new_node;
            }

            EXPORT
            val_t* lookup(hashmap_t *h, uint64_t key)
            {
                int pos = hash_code(h, key);
                node_t *list = h->list[pos];
                node_t *temp = list;
                while (temp)
                {
                    if (temp->key == key)
                    {
                        return temp->val;
                    }
                    temp = temp->next;
                }
                return h->default_value;
            }
        """
        write(helpers)

    def _create_code(self, write):
        write("#include <stdint.h>")
        write("#include <stdlib.h>")
        write("#include <string.h>")

        # windows dllexport needed to make symbols visible
        if platform.system() == "Windows":
            write("#define EXPORT __declspec(dllexport)")
        else:
            write("#define EXPORT")

        # multiplication macro
        #  for efficient 64x64 -> 128 bit multiplication without uint128_t
        #  as -O0 optimization does not handle uint128_t well
        machine_alias = {"amd64": "x86_64", "aarch64": "arm64", "aarch64_be": "arm64"}
        machine = platform.machine().lower()
        machine = machine_alias.get(machine, machine)
        mulinstr = {
            "x86_64": '"mulq %q3":"=a"(pl),"=d"(ph):"%0"(t0),"r"(t1):"cc"',
            "arm64": '"mul %0, %2, %3\\n\\t" \\\n'
            '"umulh %1, %2, %3":"=&r"(pl),"=r"(ph):"r"(t0),"r"(t1):"cc"',
            "mips64": '"dmultu %2, %3\\n\\t" \\\n'
            '"tmflo %0\\n\\t" \\\n'
            '"mfhi %1":"=r"(pl),"=r"(ph):"r"(t0),"r"(t1)',
        }
        if machine in mulinstr:
            write(f"#define mul128(t0, t1, pl, ph) __asm__({mulinstr[machine]})")

        # declare memories
        mems = {net.op_param[1] for net in self.block.logic_subset("m@")}
        for key in self._memmap:
            if key not in mems:
                msg = "unrecognized MemBlock in memory_value_map"
                raise PyrtlError(msg)
            if isinstance(key, RomBlock):
                msg = "RomBlock in memory_value_map"
                raise PyrtlError(msg)
        self._declare_mem_helpers(write)
        roms = {mem for mem in mems if isinstance(mem, RomBlock)}
        self._declare_roms(write, roms)
        mems = {
            mem
            for mem in mems
            if isinstance(mem, MemBlock) and not isinstance(mem, RomBlock)
        }
        self._declare_mems(write, mems)

        # single step function
        write("static void sim_run_step(uint64_t inputs[], uint64_t outputs[]) {")
        write("uint64_t tmp, carry, tmphi, tmplo;")  # temporary variables

        # declare wire vectors
        for w in self.block.wirevector_set:
            self._declare_wv(write, w)

        # inputs copied in
        inputs = list(self.block.wirevector_subset(Input))
        # for each input wire, start and number of elements in input array
        self._inputpos = {}
        self._inputbw = {}  # bitwidth of each input wire
        ipos = 0
        for w in inputs:
            self._inputpos[w.name] = ipos, self._limbs(w)
            self._inputbw[w.name] = w.bitwidth
            for n in range(self._limbs(w)):
                write(f"{self.varname[w]}[{n}] = inputs[{ipos}];")
                ipos += 1
        self._ibufsz = ipos  # total length of input array

        # combinational logic
        op_builders = {
            "m": self._build_memread,
            "w": self._build_wire,
            "~": self._build_not,
            "&": self._build_bitwise,
            "|": self._build_bitwise,
            "^": self._build_bitwise,
            "n": self._build_nand,
            "=": self._build_eq,
            "<": self._build_cmp,
            ">": self._build_cmp,
            "x": self._build_mux,
            "+": self._build_add,
            "-": self._build_sub,
            "*": self._build_mul,
            "c": self._build_concat,
            "s": self._build_select,
        }
        for net in self.block:  # topological order
            if net.op in "r@":
                continue  # skip synchronized nets
            op, param, args, dest = net.op, net.op_param, net.args, net.dests[0]
            write(
                "// net {op} : {args} -> {dest}".format(
                    op=op,
                    args=", ".join(self.varname[x] for x in args),
                    dest=self.varname[dest],
                )
            )
            op_builders[op](write, op, param, args, dest)

        # memory writes
        for net in self.block.logic_subset("@"):
            mem = net.op_param[1]
            write(f"if ({self.varname[net.args[2]]}[0]) {{")
            write(
                f"insert({self.varname[mem]}, {self.varname[net.args[0]]}[0], "
                f"{self.varname[net.args[1]]});"
            )
            write("}")

        # register updates
        regnets = list(self.block.logic_subset("r"))
        for x, net in enumerate(regnets):
            rin = net.args[0]
            write(f"uint64_t regtmp{x}[{self._limbs(rin)}];")
            for n in range(self._limbs(rin)):
                write(f"regtmp{x}[{n}] = {self.varname[rin]}[{n}];")
        # double loop to ensure register-to-register chains update correctly
        for x, net in enumerate(regnets):
            rout = net.dests[0]
            for n in range(self._limbs(rout)):
                write(f"{self.varname[rout]}[{n}] = regtmp{x}[{n}];")

        # output copied out
        outputs = list(self.block.wirevector_subset(Output))
        # for each output wire, start and number of elements in output array
        self._outputpos = {}
        opos = 0
        for w in outputs:
            self._outputpos[w.name] = opos, self._limbs(w)
            for n in range(self._limbs(w)):
                write(f"outputs[{opos}] = {self.varname[w]}[{n}];")
                opos += 1
        self._obufsz = opos  # total length of output array
        write("}")

        # entry point
        write("EXPORT")
        write(
            "void sim_run_all("
            "uint64_t stepcount, uint64_t inputs[], uint64_t outputs[]) {"
        )
        write("uint64_t input_pos = 0, output_pos = 0;")
        write("for (uint64_t stepnum = 0; stepnum < stepcount; stepnum++) {")
        write("sim_run_step(inputs+input_pos, outputs+output_pos);")
        write(f"input_pos += {self._ibufsz};")
        write(f"output_pos += {self._obufsz};")
        write("}}")

    def __del__(self):
        """Handle removal of the DLL when the simulator is deleted."""
        if self._dll is not None:
            handle = self._dll._handle
            if platform.system() == "Windows":
                _ctypes.FreeLibrary(handle)
            else:
                _ctypes.dlclose(handle)
            self._dll = None
        if self._dir is not None:
            shutil.rmtree(self._dir)
            self._dir = None
